// binfile - class library for files/streams - console io
// Copyright (c) 1997 Niklas Beisert
// See COPYING (GNU General Public License 2) for license

#ifndef NOUNISTD
#ifdef DOS
#include <io.h>
#endif
#include <unistd.h>
#include <fcntl.h>
#include "binfcon.h"

consolebinfile::consolebinfile()
{
  openmode(moderead|modewrite, 0, 0);
}

consolebinfile::~consolebinfile()
{
  closemode();
}

errstat consolebinfile::rawclose()
{
  return -1;
}

binfilepos consolebinfile::rawread(void *buf, binfilepos len)
{
  return ::read(STDIN_FILENO, buf, len);
}

binfilepos consolebinfile::rawwrite(const void *buf, binfilepos len)
{
  return ::write(STDOUT_FILENO, buf, len);
}

binfilepos consolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

iconsolebinfile::iconsolebinfile()
{
  openmode(moderead, 0, 0);
}

iconsolebinfile::~iconsolebinfile()
{
  closemode();
}

errstat iconsolebinfile::rawclose()
{
  return -1;
}

binfilepos iconsolebinfile::rawread(void *buf, binfilepos len)
{
  return ::read(STDIN_FILENO, buf, len);
}

binfilepos iconsolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

oconsolebinfile::oconsolebinfile()
{
  openmode(modewrite, 0, 0);
}

oconsolebinfile::~oconsolebinfile()
{
  closemode();
}

errstat oconsolebinfile::rawclose()
{
  return -1;
}

binfilepos oconsolebinfile::rawwrite(const void *buf, binfilepos len)
{
  return ::write(STDOUT_FILENO, buf, len);
}

binfilepos oconsolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

econsolebinfile::econsolebinfile()
{
  openmode(modewrite, 0, 0);
}

econsolebinfile::~econsolebinfile()
{
  closemode();
}

errstat econsolebinfile::rawclose()
{
  return -1;
}

binfilepos econsolebinfile::rawwrite(const void *buf, binfilepos len)
{
  return ::write(STDERR_FILENO, buf, len);
}

binfilepos econsolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

#else

#include <stdio.h>
#include "binfcon.h"

consolebinfile::consolebinfile()
{
  openmode(moderead|modewrite, 0, 0);
}

consolebinfile::~consolebinfile()
{
  closemode();
}

errstat consolebinfile::rawclose()
{
  return -1;
}

binfilepos consolebinfile::rawread(void *buf, binfilepos len)
{
  return fread(buf, 1, len, stdin);
}

binfilepos consolebinfile::rawwrite(const void *buf, binfilepos len)
{
  return fwrite(buf, 1, len, stdout);
}

binfilepos consolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

iconsolebinfile::iconsolebinfile()
{
  openmode(moderead, 0, 0);
}

iconsolebinfile::~iconsolebinfile()
{
  closemode();
}

errstat iconsolebinfile::rawclose()
{
  return -1;
}

binfilepos iconsolebinfile::rawread(void *buf, binfilepos len)
{
  return fread(buf, 1, len, stdin);
}

binfilepos iconsolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

oconsolebinfile::oconsolebinfile()
{
  openmode(modewrite, 0, 0);
}

oconsolebinfile::~oconsolebinfile()
{
  closemode();
}

errstat oconsolebinfile::rawclose()
{
  return -1;
}

binfilepos oconsolebinfile::rawwrite(const void *buf, binfilepos len)
{
  return fwrite(buf, 1, len, stdout);
}

binfilepos oconsolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

econsolebinfile::econsolebinfile()
{
  openmode(modewrite, 0, 0);
}

econsolebinfile::~econsolebinfile()
{
  closemode();
}

errstat econsolebinfile::rawclose()
{
  return -1;
}

binfilepos econsolebinfile::rawwrite(const void *buf, binfilepos len)
{
  return fwrite(buf, 1, len, stderr);
}

binfilepos econsolebinfile::rawioctl(intm code, void *buf, binfilepos len)
{
  switch (code)
  {
  default: return binfile::rawioctl(code, buf, len);
  }
}

#endif
